@@ -1,5 +1,5 @@
 // Copyright Contributors to the OpenImageIO project.
-// SPDX-License-Identifier: BSD-3-Clause and Apache-2.0
+// SPDX-License-Identifier: Apache-2.0
 // https://github.com/AcademySoftwareFoundation/OpenImageIO
 
 #include <cerrno>
@@ -14,11 +14,11 @@
 #include <OpenImageIO/Imath.h>
 #include <OpenImageIO/platform.h>
 
-#include <boost/version.hpp>
-
+#include <OpenEXR/ImfArray.h>
 #include <OpenEXR/ImfChannelList.h>
 #include <OpenEXR/ImfEnvmap.h>
 #include <OpenEXR/ImfInputFile.h>
+#include <OpenEXR/ImfRgba.h>
 #include <OpenEXR/ImfTestFile.h>
 #include <OpenEXR/ImfTiledInputFile.h>
 
@@ -79,187 +79,6 @@ OIIO_PRAGMA_VISIBILITY_POP
 OIIO_PLUGIN_NAMESPACE_BEGIN
 
 
-// Custom file input stream, copying code from the class StdIFStream in OpenEXR,
-// which would have been used if we just provided a filename. The difference is
-// that this can handle UTF-8 file paths on all platforms.
-class OpenEXRInputStream final : public Imf::IStream {
-public:
-    OpenEXRInputStream(const char* filename, Filesystem::IOProxy* io)
-        : Imf::IStream(filename)
-        , m_io(io)
-    {
-        if (!io || io->mode() != Filesystem::IOProxy::Read)
-            throw Iex::IoExc("File input failed.");
-    }
-    bool read(char c[], int n) override
-    {
-        OIIO_DASSERT(m_io);
-        if (m_io->read(c, n) != size_t(n))
-            throw Iex::IoExc("Unexpected end of file.");
-        return n;
-    }
-#if OIIO_USING_IMATH >= 3
-    uint64_t tellg() override { return m_io->tell(); }
-    void seekg(uint64_t pos) override
-    {
-        if (!m_io->seek(pos))
-            throw Iex::IoExc("File input failed.");
-    }
-#else
-    Imath::Int64 tellg() override { return m_io->tell(); }
-    void seekg(Imath::Int64 pos) override
-    {
-        if (!m_io->seek(pos))
-            throw Iex::IoExc("File input failed.");
-    }
-#endif
-    void clear() override {}
-
-private:
-    Filesystem::IOProxy* m_io = nullptr;
-};
-
-
-
-class OpenEXRInput final : public ImageInput {
-public:
-    OpenEXRInput();
-    ~OpenEXRInput() override { close(); }
-    const char* format_name(void) const override { return "openexr"; }
-    int supports(string_view feature) const override
-    {
-        return (feature == "arbitrary_metadata"
-                || feature == "exif"  // Because of arbitrary_metadata
-                || feature == "iptc"  // Because of arbitrary_metadata
-                || feature == "ioproxy");
-    }
-    bool valid_file(const std::string& filename) const override;
-    bool open(const std::string& name, ImageSpec& newspec,
-              const ImageSpec& config) override;
-    bool open(const std::string& name, ImageSpec& newspec) override
-    {
-        return open(name, newspec, ImageSpec());
-    }
-    bool close() override;
-    int current_subimage(void) const override { return m_subimage; }
-    int current_miplevel(void) const override { return m_miplevel; }
-    bool seek_subimage(int subimage, int miplevel) override;
-    ImageSpec spec(int subimage, int miplevel) override;
-    ImageSpec spec_dimensions(int subimage, int miplevel) override;
-    bool read_native_scanline(int subimage, int miplevel, int y, int z,
-                              void* data) override;
-    bool read_native_scanlines(int subimage, int miplevel, int ybegin, int yend,
-                               int z, void* data) override;
-    bool read_native_scanlines(int subimage, int miplevel, int ybegin, int yend,
-                               int z, int chbegin, int chend,
-                               void* data) override;
-    bool read_native_tile(int subimage, int miplevel, int x, int y, int z,
-                          void* data) override;
-    bool read_native_tiles(int subimage, int miplevel, int xbegin, int xend,
-                           int ybegin, int yend, int zbegin, int zend,
-                           void* data) override;
-    bool read_native_tiles(int subimage, int miplevel, int xbegin, int xend,
-                           int ybegin, int yend, int zbegin, int zend,
-                           int chbegin, int chend, void* data) override;
-    bool read_native_deep_scanlines(int subimage, int miplevel, int ybegin,
-                                    int yend, int z, int chbegin, int chend,
-                                    DeepData& deepdata) override;
-    bool read_native_deep_tiles(int subimage, int miplevel, int xbegin,
-                                int xend, int ybegin, int yend, int zbegin,
-                                int zend, int chbegin, int chend,
-                                DeepData& deepdata) override;
-
-    bool set_ioproxy(Filesystem::IOProxy* ioproxy) override
-    {
-        m_io = ioproxy;
-        return true;
-    }
-
-private:
-    struct PartInfo {
-        std::atomic_bool initialized;
-        ImageSpec spec;
-        int topwidth;      ///< Width of top mip level
-        int topheight;     ///< Height of top mip level
-        int levelmode;     ///< The level mode
-        int roundingmode;  ///< Rounding mode
-        bool cubeface;     ///< It's a cubeface environment map
-        int nmiplevels;    ///< How many MIP levels are there?
-        Imath::Box2i top_datawindow;
-        Imath::Box2i top_displaywindow;
-        std::vector<Imf::PixelType> pixeltype;  ///< Imf pixel type for each chan
-        std::vector<int> chanbytes;  ///< Size (in bytes) of each channel
-
-        PartInfo()
-            : initialized(false)
-        {
-        }
-        PartInfo(const PartInfo& p)
-            : initialized((bool)p.initialized)
-            , spec(p.spec)
-            , topwidth(p.topwidth)
-            , topheight(p.topheight)
-            , levelmode(p.levelmode)
-            , roundingmode(p.roundingmode)
-            , cubeface(p.cubeface)
-            , nmiplevels(p.nmiplevels)
-            , top_datawindow(p.top_datawindow)
-            , top_displaywindow(p.top_displaywindow)
-            , pixeltype(p.pixeltype)
-            , chanbytes(p.chanbytes)
-        {
-        }
-        ~PartInfo() {}
-        bool parse_header(OpenEXRInput* in, const Imf::Header* header);
-        bool query_channels(OpenEXRInput* in, const Imf::Header* header);
-        void compute_mipres(int miplevel, ImageSpec& spec) const;
-    };
-    friend struct PartInfo;
-
-    std::vector<PartInfo> m_parts;               ///< Image parts
-    OpenEXRInputStream* m_input_stream;          ///< Stream for input file
-    Imf::MultiPartInputFile* m_input_multipart;  ///< Multipart input
-    Imf::InputPart* m_scanline_input_part;
-    Imf::TiledInputPart* m_tiled_input_part;
-    Imf::DeepScanLineInputPart* m_deep_scanline_input_part;
-    Imf::DeepTiledInputPart* m_deep_tiled_input_part;
-    Filesystem::IOProxy* m_io = nullptr;
-    std::unique_ptr<Filesystem::IOProxy> m_local_io;
-    int m_subimage;                     ///< What subimage are we looking at?
-    int m_nsubimages;                   ///< How many subimages are there?
-    int m_miplevel;                     ///< What MIP level are we looking at?
-    std::vector<float> m_missingcolor;  ///< Color for missing tile/scanline
-
-    void init()
-    {
-        m_input_stream             = NULL;
-        m_input_multipart          = NULL;
-        m_scanline_input_part      = NULL;
-        m_tiled_input_part         = NULL;
-        m_deep_scanline_input_part = NULL;
-        m_deep_tiled_input_part    = NULL;
-        m_subimage                 = -1;
-        m_miplevel                 = -1;
-        m_io                       = nullptr;
-        m_local_io.reset();
-        m_missingcolor.clear();
-    }
-
-    bool valid_file(const std::string& filename, Filesystem::IOProxy* io) const;
-
-    bool read_native_tiles_individually(int subimage, int miplevel, int xbegin,
-                                        int xend, int ybegin, int yend,
-                                        int zbegin, int zend, int chbegin,
-                                        int chend, void* data, stride_t xstride,
-                                        stride_t ystride);
-
-    // Fill in with 'missing' color/pattern.
-    void fill_missing(int xbegin, int xend, int ybegin, int yend, int zbegin,
-                      int zend, int chbegin, int chend, void* data,
-                      stride_t xstride, stride_t ystride);
-};
-
-
 
 // Obligatory material to make this a recognizable imageio plugin:
 OIIO_PLUGIN_EXPORTS_BEGIN
@@ -338,24 +157,13 @@ OpenEXRInput::OpenEXRInput() { init(); }
 
 
 bool
-OpenEXRInput::valid_file(const std::string& filename) const
+OpenEXRInput::valid_file(Filesystem::IOProxy* ioproxy) const
 {
-    return valid_file(filename, nullptr);
-}
+    if (!ioproxy || ioproxy->mode() != Filesystem::IOProxy::Mode::Read)
+        return false;
 
-
-
-bool
-OpenEXRInput::valid_file(const std::string& filename,
-                         Filesystem::IOProxy* io) const
-{
     try {
-        std::unique_ptr<Filesystem::IOProxy> local_io;
-        if (!io) {
-            io = new Filesystem::IOFile(filename, Filesystem::IOProxy::Read);
-            local_io.reset(io);
-        }
-        OpenEXRInputStream IStream(filename.c_str(), io);
+        OpenEXRInputStream IStream("", ioproxy);
         return Imf::isOpenExrFile(IStream);
     } catch (const std::exception& e) {
         return false;
@@ -381,7 +189,16 @@ OpenEXRInput::open(const std::string& name, ImageSpec& newspec,
         errorf("Could not open file \"%s\"", name);
         return false;
     }
-    if (!valid_file(name, m_io)) {
+
+    // If we weren't given an IOProxy, create one now that just reads from
+    // the file.
+    if (!m_io) {
+        m_io = new Filesystem::IOFile(name, Filesystem::IOProxy::Read);
+        m_local_io.reset(m_io);
+    }
+    OIIO_ASSERT(m_io);
+
+    if (!valid_file(m_io)) {
         errorf("\"%s\" is not an OpenEXR file", name);
         return false;
     }
@@ -416,14 +233,8 @@ OpenEXRInput::open(const std::string& name, ImageSpec& newspec,
     // Clear the spec with default constructor
     m_spec = ImageSpec();
 
-    // Establish an input stream. If we weren't given an IOProxy, create one
-    // now that just reads from the file.
+    // Establish an input stream.
     try {
-        if (!m_io) {
-            m_io = new Filesystem::IOFile(name, Filesystem::IOProxy::Read);
-            m_local_io.reset(m_io);
-        }
-        OIIO_ASSERT(m_io);
         if (m_io->mode() != Filesystem::IOProxy::Read) {
             // If the proxy couldn't be opened in write mode, try to
             // return an error.
@@ -582,18 +393,10 @@ OpenEXRInput::PartInfo::parse_header(OpenEXRInput* in,
         case Imf::ZIP_COMPRESSION: comp = "zip"; break;
         case Imf::PIZ_COMPRESSION: comp = "piz"; break;
         case Imf::PXR24_COMPRESSION: comp = "pxr24"; break;
-#ifdef IMF_B44_COMPRESSION
-            // The enum Imf::B44_COMPRESSION is not defined in older versions
-            // of OpenEXR, and there are no explicit version numbers in the
-            // headers.  BUT this other related #define is present only in
-            // the newer version.
         case Imf::B44_COMPRESSION: comp = "b44"; break;
         case Imf::B44A_COMPRESSION: comp = "b44a"; break;
-#endif
-#if OPENEXR_CODED_VERSION >= 20200
         case Imf::DWAA_COMPRESSION: comp = "dwaa"; break;
         case Imf::DWAB_COMPRESSION: comp = "dwab"; break;
-#endif
         default: break;
         }
         if (comp)
@@ -928,6 +731,38 @@ suffixfound(string_view name, span<ChanNameHolder> chans)
 }
 
 
+// Returns the index of that channel name (suffix only) in the list, or -1 in case of failure.
+static int
+get_index_of_suffix(string_view name, span<ChanNameHolder> chans)
+{
+    for (size_t i = 0, n = chans.size(); i < n; ++i)
+        if (Strutil::iequals(name, chans[i].suffix))
+            return static_cast<int>(i);
+    return -1;
+}
+
+
+// Is this a luminance-chroma image, i.e., Y/BY/RY or Y/BY/RY/A or Y/BY/RY/Alpha?
+//
+// Note that extra channels are not supported.
+static bool
+is_luminance_chroma(span<ChanNameHolder> chans)
+{
+    if (chans.size() < 3 || chans.size() > 4)
+        return false;
+    if (!suffixfound("Y", chans))
+        return false;
+    if (!suffixfound("BY", chans))
+        return false;
+    if (!suffixfound("RY", chans))
+        return false;
+    if (chans.size() == 4 && !suffixfound("A", chans)
+        && !suffixfound("Alpha", chans))
+        return false;
+    return true;
+}
+
+
 }  // namespace
 
 
@@ -937,10 +772,8 @@ OpenEXRInput::PartInfo::query_channels(OpenEXRInput* in,
                                        const Imf::Header* header)
 {
     OIIO_DASSERT(!initialized);
-    bool ok        = true;
-    spec.nchannels = 0;
+    bool ok = true;
     const Imf::ChannelList& channels(header->channels());
-    std::vector<std::string> channelnames;  // Order of channels in file
     std::vector<ChanNameHolder> cnh;
     int c = 0;
     for (auto ci = channels.begin(); ci != channels.end(); ++c, ++ci)
@@ -987,6 +820,34 @@ OpenEXRInput::PartInfo::query_channels(OpenEXRInput* in,
     // Now we should have cnh sorted into the order that we want to present
     // to the OIIO client.
 
+    // Limitations for luminance-chroma images: no tiling, no deep samples, no
+    // miplevels/subimages, no extra channels.
+    luminance_chroma = is_luminance_chroma(cnh);
+    if (luminance_chroma) {
+        spec.attribute("openexr:luminancechroma", 1);
+        spec.format    = TypeDesc::HALF;
+        spec.nchannels = cnh.size();
+        if (spec.nchannels == 3) {
+            spec.channelnames  = { "R", "G", "B" };
+            spec.alpha_channel = -1;
+            spec.z_channel     = -1;
+        } else {
+            OIIO_ASSERT(spec.nchannels == 4);
+            int index_a = get_index_of_suffix("A", cnh);
+            if (index_a != -1) {
+                spec.channelnames  = { "R", "G", "B", "A" };
+                spec.alpha_channel = index_a;
+            } else {
+                spec.channelnames  = { "R", "G", "B", "Alpha" };
+                spec.alpha_channel = get_index_of_suffix("Alpha", cnh);
+                OIIO_ASSERT(spec.alpha_channel != -1);
+            }
+            spec.z_channel = -1;
+        }
+        spec.channelformats.clear();
+        return true;
+    }
+
     spec.format         = TypeDesc::UNKNOWN;
     bool all_one_format = true;
     for (int c = 0; c < spec.nchannels; ++c) {
@@ -1009,7 +870,8 @@ OpenEXRInput::PartInfo::query_channels(OpenEXRInput* in,
             in->errorfmt(
                 "Subsampled channels are not supported (channel \"{}\" has sampling {},{}).",
                 cnh[c].fullname, cnh[c].xSampling, cnh[c].ySampling);
-            // FIXME: Some day, we should handle channel subsampling.
+            // FIXME: Some day, we should handle channel subsampling (beyond the luminance chroma
+            // special case, possibly replacing it).
         }
     }
     OIIO_DASSERT((int)spec.channelnames.size() == spec.nchannels);
@@ -1106,8 +968,18 @@ OpenEXRInput::seek_subimage(int subimage, int miplevel)
         m_deep_scanline_input_part = NULL;
         delete m_deep_tiled_input_part;
         m_deep_tiled_input_part = NULL;
+        delete m_input_rgba;
+        m_input_rgba = NULL;
         try {
-            if (part.spec.deep) {
+            if (part.luminance_chroma) {
+                if (subimage != 0 || miplevel != 0) {
+                    errorf(
+                        "Non-zero subimage or miplevel are not supported for luminance-chroma images.");
+                    return false;
+                }
+                m_input_stream->seekg(0);
+                m_input_rgba = new Imf::RgbaInputFile(*m_input_stream);
+            } else if (part.spec.deep) {
                 if (part.spec.tile_width)
                     m_deep_tiled_input_part
                         = new Imf::DeepTiledInputPart(*m_input_multipart,
@@ -1130,6 +1002,7 @@ OpenEXRInput::seek_subimage(int subimage, int miplevel)
             m_tiled_input_part         = NULL;
             m_deep_scanline_input_part = NULL;
             m_deep_tiled_input_part    = NULL;
+            m_input_rgba               = NULL;
             return false;
         } catch (...) {  // catch-all for edge cases or compiler bugs
             errorf("OpenEXR exception: unknown");
@@ -1137,6 +1010,7 @@ OpenEXRInput::seek_subimage(int subimage, int miplevel)
             m_tiled_input_part         = NULL;
             m_deep_scanline_input_part = NULL;
             m_deep_tiled_input_part    = NULL;
+            m_input_rgba               = NULL;
             return false;
         }
     }
@@ -1218,6 +1092,7 @@ OpenEXRInput::close()
     delete m_tiled_input_part;
     delete m_deep_scanline_input_part;
     delete m_deep_tiled_input_part;
+    delete m_input_rgba;
     delete m_input_stream;
     init();  // Reset to initial state
     return true;
@@ -1244,7 +1119,6 @@ OpenEXRInput::read_native_scanlines(int subimage, int miplevel, int ybegin,
 }
 
 
-
 bool
 OpenEXRInput::read_native_scanlines(int subimage, int miplevel, int ybegin,
                                     int yend, int /*z*/, int chbegin, int chend,
@@ -1256,11 +1130,6 @@ OpenEXRInput::read_native_scanlines(int subimage, int miplevel, int ybegin,
     chend = clamp(chend, chbegin + 1, m_spec.nchannels);
     //    std::cerr << "openexr rns " << ybegin << ' ' << yend << ", channels "
     //              << chbegin << "-" << (chend-1) << "\n";
-    if (!m_scanline_input_part) {
-        errorf(
-            "called OpenEXRInput::read_native_scanlines without an open file");
-        return false;
-    }
 
     // Compute where OpenEXR needs to think the full buffers starts.
     // OpenImageIO requires that 'data' points to where the client wants
@@ -1273,6 +1142,51 @@ OpenEXRInput::read_native_scanlines(int subimage, int miplevel, int ybegin,
     char* buf = (char*)data - m_spec.x * pixelbytes - ybegin * scanlinebytes;
 
     try {
+        if (part.luminance_chroma) {
+            Imath::Box2i dw = m_input_rgba->dataWindow();
+            if (dw.min.x != 0 || dw.min.y != 0
+                || dw != m_input_rgba->displayWindow()) {
+                errorf(
+                    "Non-trivial data and/or display windows are not supported for luminance-chroma images.");
+                return false;
+            }
+            int dw_width     = dw.max.x - dw.min.x + 1;
+            int dw_height    = dw.max.y - dw.min.y + 1;
+            int chunk_height = yend - ybegin;
+            // FIXME Are these assumptions correct?
+            OIIO_ASSERT(ybegin >= dw.min.y);
+            OIIO_ASSERT(yend <= dw.max.y + 1);
+            OIIO_ASSERT(chunk_height <= dw_height);
+
+            Imf::Array2D<Imf::Rgba> pixels(chunk_height, dw_width);
+            m_input_rgba->setFrameBuffer(&pixels[0][0] - dw.min.x
+                                             - ybegin * dw_width,
+                                         1, dw_width);
+            m_input_rgba->readPixels(ybegin, yend - 1);
+
+            // FIXME There is probably some optimized code for this somewhere.
+            for (int c = chbegin; c < chend; ++c) {
+                size_t chanbytes = m_spec.channelformat(c).size();
+                half* src        = &pixels[0][0].r + c;
+                half* dst        = (half*)((char*)data + c * chanbytes);
+                for (int y = ybegin; y < yend; ++y) {
+                    for (int x = 0; x < m_spec.width; ++x) {
+                        *dst = *src;
+                        src += 4;  // always advance 4 RGBA halfs
+                        dst += m_spec.nchannels;
+                    }
+                }
+            }
+
+            return true;
+        }
+
+        if (!m_scanline_input_part) {
+            errorf(
+                "called OpenEXRInput::read_native_scanlines without an open file");
+            return false;
+        }
+
         Imf::FrameBuffer frameBuffer;
         size_t chanoffset = 0;
         for (int c = chbegin; c < chend; ++c) {
@@ -1338,6 +1252,12 @@ OpenEXRInput::read_native_tiles(int subimage, int miplevel, int xbegin,
     if (!seek_subimage(subimage, miplevel))
         return false;
     chend = clamp(chend, chbegin + 1, m_spec.nchannels);
+    const PartInfo& part(m_parts[m_subimage]);
+    if (part.luminance_chroma) {
+        errorf(
+            "OpenEXRInput::read_native_tiles is not supported for luminance-chroma images");
+        return false;
+    }
 #if 0
     std::cerr << "openexr rnt " << xbegin << ' ' << xend << ' ' << ybegin 
               << ' ' << yend << ", chans " << chbegin
@@ -1354,7 +1274,6 @@ OpenEXRInput::read_native_tiles(int subimage, int miplevel, int xbegin,
     // to put the pixels being read, but OpenEXR's frameBuffer.insert()
     // wants where the address of the "virtual framebuffer" for the
     // whole image.
-    const PartInfo& part(m_parts[m_subimage]);
     size_t pixelbytes = m_spec.pixel_bytes(chbegin, chend, true);
     int firstxtile    = (xbegin - m_spec.x) / m_spec.tile_width;
     int firstytile    = (ybegin - m_spec.y) / m_spec.tile_height;
@@ -1510,6 +1429,12 @@ OpenEXRInput::read_native_deep_scanlines(int subimage, int miplevel, int ybegin,
     lock_guard lock(*this);
     if (!seek_subimage(subimage, miplevel))
         return false;
+    const PartInfo& part(m_parts[m_subimage]);
+    if (part.luminance_chroma) {
+        errorf(
+            "OpenEXRInput::read_native_deep_scanlines is not supported for luminance-chroma images");
+        return false;
+    }
     if (m_deep_scanline_input_part == NULL) {
         errorf(
             "called OpenEXRInput::read_native_deep_scanlines without an open file");
@@ -1517,7 +1442,6 @@ OpenEXRInput::read_native_deep_scanlines(int subimage, int miplevel, int ybegin,
     }
 
     try {
-        const PartInfo& part(m_parts[m_subimage]);
         size_t npixels = (yend - ybegin) * m_spec.width;
         chend          = clamp(chend, chbegin + 1, m_spec.nchannels);
         int nchans     = chend - chbegin;
@@ -1582,6 +1506,12 @@ OpenEXRInput::read_native_deep_tiles(int subimage, int miplevel, int xbegin,
     lock_guard lock(*this);
     if (!seek_subimage(subimage, miplevel))
         return false;
+    const PartInfo& part(m_parts[m_subimage]);
+    if (part.luminance_chroma) {
+        errorf(
+            "OpenEXRInput::read_native_deep_tiles is not supported for luminance-chroma images");
+        return false;
+    }
     if (m_deep_tiled_input_part == NULL) {
         errorf(
             "called OpenEXRInput::read_native_deep_tiles without an open file");
@@ -1589,7 +1519,6 @@ OpenEXRInput::read_native_deep_tiles(int subimage, int miplevel, int xbegin,
     }
 
     try {
-        const PartInfo& part(m_parts[m_subimage]);
         size_t width   = xend - xbegin;
         size_t height  = yend - ybegin;
         size_t npixels = width * height;
